unit ULogic;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, math, Graphics;

const
  Train = 'Naja, da muss wohl noch ein wenig geübt werden ;)';
  Super = 'Gar nicht schlecht, eine wirklich gute Leistung!';
  Great = 'Unglaublich, dass Du das geschafft hast! TipTop!';
  Boring = 'Ist Dir irgendwie langweilig?';

type
  TOrientation = (orDown, orLeft, orRight, orUp);

  TSnakePart = record
    Orientation : TOrientation;
    x: Integer;
    y: Integer;
  end;

  TNibbles = record
    SpareTime: Integer;
    x: Integer;
    y: Integer;
    ImgIndex: Integer;
  end;

  { TFood }

  TFood = class(TObject)
  private
    FNibbles: array of TNibbles;
    procedure Delete(Index: Integer);
  public
    procedure Add(P: TPoint);
    procedure Welk;
    procedure Clear;
    function IsOnPoint(P: TPoint; out Index: Integer): Boolean;
    function Eat(P: TPoint): Boolean;
    function GetColor(Index: Integer): TColor;
    function GetSpareTimeSize(Index: Integer): Byte;
    function GetImageIndex(Index: Integer): Integer;
  end;

  { TSnake }

  TSnake = class(TObject)
  private
    FFeldMaxX: Integer;
    FFeldMaxY: Integer;
    FOrientation: TList;
    FOldOrientation: TOrientation;
    FSnakeParts: array of TSnakePart;
    FScore: Integer;
    function GetHead: TPoint;
    function GetOrientation: TOrientation;
    function GetOrientationSnakePart(Index: Integer): TOrientation;
    procedure SetOrientation(AOrientation: TOrientation);
  public
    constructor Create;
    destructor Destroy; override;
    function IndexOf(P: TPoint): Integer;
    function GetColor(Index: Integer): TColor;
    function GetImageIndex(Index: Integer): Integer;
    procedure Init;
    procedure Move;
    procedure IncScore(Credits: Integer);
    procedure Grow;
    function Collision: Boolean;
    function IsOnPoint(P: TPoint; MitKopf: Boolean = True): Boolean;
    property Head: TPoint read GetHead;
    property Orientation: TOrientation read GetOrientation write SetOrientation;
    property FeldMaxX: Integer read FFeldMaxX write FFeldMaxX;
    property FeldMaxY: Integer read FFeldMaxY write FFeldMaxY;
    property Score: Integer read FScore;
  end;

implementation

{ TFood }

procedure TFood.Delete(Index: Integer);
var
  i: Integer;
begin
  for i := Index to High(FNibbles) do
    FNibbles[i] := FNibbles[i + 1];
  SetLength(FNibbles, Length(FNibbles) - 1);
end;

procedure TFood.Add(P: TPoint);
begin
  SetLength(FNibbles, Length(FNibbles) + 1);
  FNibbles[High(FNibbles)].SpareTime := 100;
  FNibbles[High(FNibbles)].ImgIndex := Random(4) + 20;
  FNibbles[High(FNibbles)].x := P.x;
  FNibbles[High(FNibbles)].y := P.y;
end;

procedure TFood.Welk;
var
  i: Integer;
begin
  for i := High(FNibbles) downto 0 do begin
    Dec(FNibbles[i].SpareTime);
    if FNibbles[i].SpareTime = 0 then
      Delete(i);
  end;
end;

procedure TFood.Clear;
var
  i: Integer;
begin
  for i := High(FNibbles) downto 0 do Delete(i);
end;

function TFood.IsOnPoint(P: TPoint; out Index: Integer
  ): Boolean;
begin
  Result := False;
  Index := 0;
  while Index < Length(FNibbles) do
    if (FNibbles[Index].x = P.x) and (FNibbles[Index].y = P.y) then begin
      Result := True;
      Exit;
    end else
      Inc(Index);
  Result := False;
  Index  := -1;
end;

function TFood.Eat(P: TPoint): Boolean;
var
  i: Integer;
begin
  if not IsOnPoint(P, i) then Exit(False);
  Delete(i);
  Result := True;
end;

function TFood.GetColor(Index: Integer): TColor;
var
  Part: Byte;
begin
  Part := 255 - GetSpareTimeSize(Index);
  Result := RGBToColor(Part, 128 + Part div 2, Part);
end;

function TFood.GetSpareTimeSize(Index: Integer): Byte;
begin
  Result := Min(FNibbles[Index].SpareTime * 10, 255);
end;

function TFood.GetImageIndex(Index: Integer): Integer;
begin
  Result := FNibbles[Index].ImgIndex;
end;

{ TSnake }

function TSnake.GetHead: TPoint;
begin
  Result := Point(FSnakeParts[0].x, FSnakeParts[0].y);
end;

function TSnake.GetOrientation: TOrientation;

  function GetFirstOrientation: TOrientation;
  begin
    if FOrientation.Count = 0 then Exit(orRight);
    Result := TOrientation({%H-}Integer(FOrientation.Items[0]));
    if FOrientation.Count = 1 then Exit;
    Result := TOrientation({%H-}Integer(FOrientation.Items[1]));
    FOrientation.Delete(0);
  end;

begin
  Result := GetFirstOrientation;
  FOldOrientation := Result;
end;

function TSnake.GetOrientationSnakePart(Index: Integer
  ): TOrientation;
begin
  Result := FSnakeParts[Index].Orientation;
end;

procedure TSnake.SetOrientation(AOrientation: TOrientation);
begin
  FOrientation.Add({%H-}Pointer(Integer(AOrientation)));
end;

constructor TSnake.Create;
begin
  inherited Create;
  FOrientation := TList.Create;
  Init;
end;

destructor TSnake.Destroy;
begin
  FreeAndNil(FOrientation);
  inherited Destroy;
end;

function TSnake.IndexOf(P: TPoint): Integer;
var
  i: Integer;
begin
  for i := 0 to High(FSnakeParts) do
    if (FSnakeParts[i].x = P.x) and (FSnakeParts[i].y = P.y) then
      Exit(i);
  Result := -1;
end;

function TSnake.GetColor(Index: Integer): TColor;
begin
  if Index = 0 then Exit(clRed);
  Result := RGBToColor(240, 120 ,50 + Trunc(Sin(Index) * 50));
end;

function TSnake.GetImageIndex(Index: Integer): Integer;

   function OrientationFromTo(orFrom, orTo: TOrientation): Integer;
   begin
     Result := -1;
     case orFrom of
       orRight: if orTo = orUp    then Result := 18
                                  else Result := 17;
       orLeft:  if orTo = orUp    then Result := 19
                                  else Result := 16;
       orDown:  if orTo = orRight then Result := 19
                                  else Result := 18;
       orUp:    if orTo = orRight then Result := 16
                                  else Result := 17;
     end;
   end;

begin
  if (Index > High(FSnakeParts)) or (Index < 0) then Exit(-1);
  if Index = 0 then
    Exit(Integer(GetOrientationSnakePart(Index)));
  if Index = High(FSnakeParts) then
    Exit(12 + Integer(GetOrientationSnakePart(Index - 1)));
  Result := 4 + (FSnakeParts[Index].x + FSnakeParts[Index].y) mod 2 * 4 +
    Integer(GetOrientationSnakePart(Index));
  if (GetOrientationSnakePart(Index) = GetOrientationSnakePart(Index - 1)) then Exit;
  Result := OrientationFromTo(GetOrientationSnakePart(Index), GetOrientationSnakePart(Index - 1));
end;

procedure TSnake.Init;
var
  i: Integer;
begin
  SetLength(FSnakeParts, 0);
  FOldOrientation := orRight;
  for i := 0 to 2 do Grow;
  FSnakeParts[0].x := 0;
  FSnakeParts[0].y := 0;
  FOrientation.Clear;
  Orientation := orRight;
  FScore   := 0;
end;

procedure TSnake.Move;
var
  i: Integer;
begin
  for i := High(FSnakeParts) downto 1 do
    FSnakeParts[i] := FSnakeParts[i - 1];
  case Orientation of
    orLeft:  FSnakeParts[0].x := FSnakeParts[0].x - 1;
    orRight: FSnakeParts[0].x := FSnakeParts[0].x + 1;
    orUp:    FSnakeParts[0].y := FSnakeParts[0].y - 1;
    orDown:  FSnakeParts[0].y := FSnakeParts[0].y + 1;
  end;
  FSnakeParts[0].Orientation := FOldOrientation;
end;

procedure TSnake.IncScore(Credits: Integer);
begin
  FScore := FScore + Credits + Length(FSnakeParts) div 10;
end;

procedure TSnake.Grow;
begin
  SetLength(FSnakeParts, Length(FSnakeParts) + 1);
  FSnakeParts[High(FSnakeParts)].x := -1;
  FSnakeParts[High(FSnakeParts)].y := -1;
  FSnakeParts[High(FSnakeParts)].Orientation := FOldOrientation;
end;

function TSnake.Collision: Boolean;
begin
  if IsOnPoint(Head, False) then Exit(True);
  if FSnakeParts[0].x < 0 then Exit(True);
  if FSnakeParts[0].y < 0 then Exit(True);
  if FSnakeParts[0].x > FeldMaxX then Exit(True);
  if FSnakeParts[0].y > FeldMaxY then Exit(True);
  Result := False;
end;

function TSnake.IsOnPoint(P: TPoint; MitKopf: Boolean
  ): Boolean;
var
  i: Integer;
begin
  Result := false;
  if MitKopf then
    i := 0
  else
    i := 1;
  while i < Length(FSnakeParts) do
    if (FSnakeParts[i].x = P.x) and (FSnakeParts[i].y = P.y) then
      Exit(true)
    else
      Inc(i);
end;

end.



