unit Unit1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls,
  Graphics, Dialogs, Grids, ExtCtrls, StdCtrls, LCLType,
  ULogic;

type

  { TfrmMain }

  TfrmMain = class(TForm)
    btnNewGame: TButton;
    btnPause: TButton;
    imgSource: TImageList;
    imgGame: TImageList;
    PlayField: TDrawGrid;
    timSpeed: TTimer;
    timPlayStep: TTimer;
    timFood: TTimer;
    procedure btnNewGameClick({%H-}Sender: TObject);
    procedure btnPauseClick({%H-}Sender: TObject);
    procedure PlayFieldDrawCell({%H-}Sender: TObject; aCol,
      aRow: Integer; aRect: TRect; {%H-}aState: TGridDrawState);
    procedure PlayFieldEnter({%H-}Sender: TObject);
    procedure FormChangeBounds({%H-}Sender: TObject);
    procedure FormCreate({%H-}Sender: TObject);
    procedure FormDestroy({%H-}Sender: TObject);
    procedure FormKeyDown({%H-}Sender: TObject; var Key: Word;
      {%H-}Shift: TShiftState);
    procedure timSpeedTimer({%H-}Sender: TObject);
    procedure timPlayStepTimer({%H-}Sender: TObject);
    procedure timFoodTimer({%H-}Sender: TObject);
  private
    Snake: TSnake;
    Food: TFood;
    procedure InitGame({%H-}Data: PtrInt);
    procedure GameOver;
  end;

var
  frmMain: TfrmMain;

implementation

{$R *.lfm}

{ TfrmMain }

procedure TfrmMain.FormChangeBounds(Sender: TObject);
begin
  Application.QueueAsyncCall(@InitGame, 0);
end;

procedure TfrmMain.PlayFieldDrawCell(Sender: TObject; aCol,
  aRow: Integer; aRect: TRect; aState: TGridDrawState);
var
  aIndex: Integer;

  function NewRecSize(Size: Byte): TRect;
  var
    aWidth, aHeight, bWidth, bHeight: Integer;
  begin
    aWidth  := aRect.Right  - aRect.Left;
    aHeight := aRect.Bottom - aRect.Top;
    bWidth  := aWidth  * Size div 255;
    bHeight := aHeight * Size div 255;
    Result.Left   := aRect.Left  + (aWidth  - bWidth)  div 2;
    Result.Top    := aRect.Top   + (aHeight - bHeight) div 2;
    Result.Right  := Result.Left + bWidth;
    Result.Bottom := Result.Top  + bHeight;
  end;

begin
  if imgGame.Count = 0 then Exit;
//  PlayField.Canvas.Brush.Color := clWhite;
//  PlayField.Canvas.FillRect(aRect);
  if Snake.IsOnPoint(Point(aCol, aRow)) then begin
    aIndex := Snake.IndexOf(Point(aCol, aRow));
//    PlayField.Canvas.Brush.Color := Snake.GetColor(aIndex);
//    PlayField.Canvas.FillRect(aRect);
    aIndex := Snake.GetImageIndex(aIndex);
    if aIndex >= 0 then
      imgGame.Draw(PlayField.Canvas, aRect.Left, aRect.Top, aIndex);
    Exit;
  end;
  if Food.IsOnPoint(Point(aCol, aRow), aIndex) then begin
//    PlayField.Canvas.Brush.Color := Food.GetColor(aIndex);
//    PlayField.Canvas.FillRect(aRect);
    aRect := NewRecSize(Food.GetSpareTimeSize(aIndex));
    aIndex := Food.GetImageIndex(aIndex);
    if aIndex >= 0 then
      imgGame.StretchDraw(PlayField.Canvas, aIndex, aRect);
  end;
end;

procedure TfrmMain.PlayFieldEnter(Sender: TObject);
begin
  btnPause.SetFocus;
end;

procedure TfrmMain.btnNewGameClick(Sender: TObject);
begin
  InitGame(0);
end;

procedure TfrmMain.btnPauseClick(Sender: TObject);
begin
  timPlayStep.Enabled    := not timPlayStep.Enabled;
  timFood.Enabled          := timPlayStep.Enabled;
  timSpeed.Enabled := timPlayStep.Enabled;
end;

procedure TfrmMain.FormCreate(Sender: TObject);
begin
  Randomize;
  Snake := TSnake.Create;
  Food   := TFood.Create;
  Left   := Screen.Width div 4;
  Top    := Screen.Height div 4;
  Width  := Screen.Width div 2;
  Height := Screen.Height div 2;
  Constraints.MinHeight := Screen.Height div 4;
  Constraints.MinWidth  := Constraints.MinHeight;
  PlayField.DefaultColWidth  := Screen.Width div 50;
  PlayField.DefaultRowHeight := PlayField.DefaultColWidth;
  Application.QueueAsyncCall(@InitGame, 0);
end;

procedure TfrmMain.FormDestroy(Sender: TObject);
begin
  Food.Free;
  Snake.Free;
end;

procedure TfrmMain.FormKeyDown(Sender: TObject;
  var Key: Word; Shift: TShiftState);
begin
  case Key of
    VK_LEFT,  VK_A: Snake.Orientation := orLeft;
    VK_RIGHT, VK_D: Snake.Orientation := orRight;
    VK_UP,    VK_W: Snake.Orientation := orUp;
    VK_DOWN,  VK_S: Snake.Orientation := orDown;
    VK_SPACE      : btnPauseClick(nil);
    else Exit;
  end;
  Key := 0;
end;

procedure TfrmMain.timSpeedTimer(Sender: TObject);
begin
  timPlayStep.Interval := timPlayStep.Interval - 1;
end;

procedure TfrmMain.timPlayStepTimer(Sender: TObject);
begin
  Snake.Move;
  if Snake.Collision then
    GameOver;
  if Food.Eat(Snake.Head) then begin
    Snake.IncScore(10);
    Snake.Grow;
  end;
  Snake.IncScore(1);
  Repaint;
  Caption := 'Gib der Schlange Futter - Punkte: ' + IntToStr(Snake.Score);
end;

procedure TfrmMain.timFoodTimer(Sender: TObject);
var
  aPoint: TPoint;
begin
  Food.Welk;
  if Random(10) = 0 then begin
    repeat
      aPoint.x := Random(PlayField.ColCount);
      aPoint.y := Random(PlayField.RowCount);
    until not Snake.IsOnPoint(aPoint);
    Food.Add(aPoint);
  end;
end;

procedure TfrmMain.InitGame(Data: PtrInt);

  procedure CenterPlayField;
  begin
    PlayField.ColCount := (ClientWidth - 20) div PlayField.DefaultColWidth;
    Snake.FeldMaxX  := PlayField.ColCount - 1;
    PlayField.RowCount := (ClientHeight - btnPause.Height - 30) div PlayField.DefaultRowHeight;
    Snake.FeldMaxY  := PlayField.RowCount - 1;
    PlayField.Width  := PlayField.DefaultColWidth  * PlayField.ColCount;
    PlayField.Height := PlayField.DefaultRowHeight * PlayField.RowCount;
    PlayField.Left := (ClientWidth - PlayField.Width) div 2;
    PlayField.Top := (ClientHeight - btnPause.Height - 30 - PlayField.Height) div 2 + 20 + btnPause.Height;
  end;
  procedure InitGameImages;
  var
    i: Integer;
    Bitmap: TBitmap;
  begin
    imgGame.Width  := PlayField.DefaultColWidth;
    imgGame.Height := PlayField.DefaultRowHeight;
    Bitmap := TBitmap.Create;
    Bitmap.Width  := imgGame.Width;
    Bitmap.Height := imgGame.Height;
    Bitmap.Canvas.Clear;
    for i := 0 to imgSource.Count - 1 do begin
      Bitmap.Canvas.Clear;
      imgSource.StretchDraw(Bitmap.Canvas, i, Rect(0, 0, Bitmap.Width, Bitmap.Height));
      imgGame.AddMasked(Bitmap, clNone);
    end;
    FreeAndNil(Bitmap);
  end;

begin
  CenterPlayField;
  if imgGame.Count = 0 then InitGameImages;
  btnPause.Width := ClientWidth div 2 - 15;
  Snake.Init;
  Food.Clear;
  timPlayStep.Enabled := False;
  timFood.Enabled     := False;
  timSpeed.Enabled    := False;
  timPlayStep.Interval := 250;
  Caption := 'Gib der Schlange Futter - Spielstart mit [Space]';
  Repaint;
end;

procedure TfrmMain.GameOver;
var
  Inhalt: String;
begin
  timPlayStep.Enabled := False;
  if Snake.Score < 1000 then Inhalt := Train
  else if Snake.Score < 5000 then Inhalt := Super
  else if Snake.Score < 20000 then Inhalt := Great
  else Inhalt := Boring;
  case QuestionDLG(IntToStr(Snake.Score) + ' Punkte',
      Inhalt + LineEnding + LineEnding + 'Willst Du ein neues Spiel beginnen?',
      mtConfirmation,
      [mrYes, 'Ja', mrNo, 'Nein'], 0) of
    mrYes: Application.QueueAsyncCall(@InitGame, 0);
    mrNo:  Close;
  end;
end;

end.

